#
# rulesdoc.mk
# GNUMake rules for processing documents
# This is like "rules.mk", but for documents.  The rules for processing
# documents are sufficiently complicated that they get there own file.
#

# Because tex and siblings produce a lot of screen output, we echo
# large obnoxious messages when invoking each command.
DELIM1="======================  "
DELIM2="  ======================"

##############################################################################
#
#			tex,latex,bibtex support
#
##############################################################################

# The interactions between bibtex and tex are complicated: from scratch,
# it requires 3 tex runs to get everything complete!  It goes like this:
# 1. tex to get required bibs listed in aux file.	(tex --> aux)
# 2. bibtex to read aux file and produce bbl file.	(aux&bib --> bbl)
# 3. tex to read bbl contents into aux file.		(bbl --> aux)
# 4. tex to get citations correct			(aux --> dvi).
# The complex set of commands and dependencies should perform all
# four of the above steps when invoked in a clean directory with "all" target.
# Note however, that we dont want any dependancy such that the bbl file
# is updated everytime the tex or aux file changes; instead the user
# must manual invoke make with the "bib" target for that.


TEXCMD=latex
%.ps: %.ms
	psroff -t -ms $< > $@

%.ps: %.man
	psroff -t -man $< > $@

%.dvi: %.tex
	@echo "$(DELIM1)$< --> $@$(DELIM2)"
	$(TEXCMD) $<
	@egrep 'Warning: ((Citation|Reference).*undefined|Rerun to get)' $*.log || exit 0 ; \
	echo "$(DELIM1)Repeat $< --> $@$(DELIM2)" && $(TEXCMD) $<

%.aux: %.tex
	@echo "$(DELIM1)$< --> $@$(DELIM2)"
	touch $*.bbl && $(TEXCMD) $<

ifdef BblFile
# This next one is tricky.  If we build the bbl file, we want to
# first make sure the aux file has been built.  We could do this
# using a dependancy "%.bbl: %.aux", which would force the aux file.
# But this would force the bbl to be rebuilt everytime the aux changed,
# which we DONT want (its too tedious).  So we dont list an explicit
# depend for the bbl file.
$(BblFile): $(BibFiles)
	if [ ! -f $(DocName).aux ] ; then $(MAKE) $(DocName).aux ; fi
	@echo "$(DELIM1)Update $@$(DELIM2)"
	bibtex $(DocName)
endif

%.ps: %.dvi
	@echo "$(DELIM1) $< --> $@ $(DELIM2)"
	@-touch /imports/moby/moby_a/foo > /dev/null 2>&1
	mydvips $< > $@
# either use dvips (user) or mydvips (prog)


tex:		FORCE
		rm -f $(DocName).dvi
		$(MAKE) $(DocName).dvi

bib:		FORCE
		rm -f $(DocName).bbl
		$(MAKE) $(DocName).bbl

spell:		FORCE
		detex $(DocName).tex | spell


##############################################################################
#
#			texinfo, latexinfo support
#
##############################################################################

ifdef InfoFile
menus:		FORCE
		$(BatchEmacs) $(DocName).tex \
		-f latexinfo-update-every -f save-buffer -f save-some-buffers \
		-kill

submenus:	FORCE
		$(BatchEmacs) $(SubDocFiles) -f latexinfo-every-node-update \
	  	-f latexinfo-all-menus-update -f save-buffer -kill

verify:		FORCE
		$(BatchEmacs) $(InfoFile) -f Info-validate

$(InfoFile):
		$(BatchEmacs) $(DocName).tex -f latexinfo-format-buffer \
		-f save-buffer -kill

info:		FORCE
		-rm -f $(InfoFile)
		$(MAKE) $(InfoFile)

tkinfo:		$(InfoFile)
		tkinfo -infofile $(InfoFile) >> errs 2>&1 &

endif


##############################################################################
#
#			nroff,troff support
#
##############################################################################

%.ps: %.rof
	psroff -t $< > $@

%.ps: %.ms
	psroff -t -ms $< > $@

#spell:
#	detex $(DOC).tex | spell

##############################################################################
#
#			generic
#
##############################################################################

%.xdvi: %.dvi
		@-touch /imports/moby/moby_a/foo > /dev/null 2>&1
		myxdvi $< >> errs 2>&1 &

xdvi:		$(DocName).xdvi

%.gview: %.ps
		ghostview $< >> errs 2>&1 &

gv gview:	%(DocName).gview

neat_clean:	FORCE
		-rm -f $(MAKEFILE).old errs *.aux *.tex~ \
		  $(DocName).log $(DocName).bbl $(DocName).blg $(DocName).toc \
		  $(DocName).cp $(DocName).fn $(DocName).ky \
		  $(DocName).pg $(DocName).tp $(DocName).vr

clean:		neat_clean
		-rm -f $(DocName).info $(InfoFile) $(DocName).dvi $(DocName).ps

# Kill all archs
pristine:	clean
		@echo
