/* 
    Author: Mei-Tjng Huang & Kennard White

    Copyright (c) 1990-93 The Regents of the University of California.
    All rights reserved.  See the file "./copyright.h" for full copyright
    notice, limitation of liability, and disclaimer of warranty provisions.
#include "copyright.h"				COPYRIGHTENDKEY
*/
#if !defined(lint) && !defined(SABER)
static char rcsid[] = "$Header: /vol/cooley/cooley1/kennard/src/eem/RCS/pfesilar.c,v 1.20 93/08/27 14:29:34 kennard Exp $";
#endif

#include "topFixup.h"
#include <stdio.h>
#include <math.h>

#include "topStd.h"
#include "topMem.h"
#include "topVec.h"

#include "eemDflt.h"
#include "eemCplx.h"
#include "eemSilar.h"
#include "eemPfe.h"

extern char _EemPfeSubPackage[];
#define SPKG _EemPfeSubPackage 

/**
  Evaluates a rational polynomal at location {pVal}.
  The numerator of the polynomial is {pPR} and and the denominator
  is {pSG}.  The result is stored in {pOut}.
  This function is called by eemEvalPolySilarA0.
**/
static void
_eemEvalPolySilar(EEMPolyCplx *pNumerPoly, EEMSilarGrp *pDenomSilars,
  EEMComplex *pIn, EEMComplex *pOut) {
    EEMComplex	denom;

    eemPolyCplxEval(pNumerPoly, pIn, pOut);
    if ( eemSilarGrpEvalPart( pDenomSilars, pIn, &denom, -1) > 0 )
        eemCplxDiv2(pOut, &denom);
}

/**
  Evaluates a fraction such that the numerator is a complex polynomial
  and the denominator is determined from the silar group.  
  Similar to the function above, the fraction is evaluated at the pole
  location of the ith silar, and in the evaluation of the fraction,
  the ith silar is removed in the calculation (done by setting the
  power of the silar to zero)
  The power of the ith silar is returned to its original value at the
  end of the function
**/
static void 
_eemEvalPolySilarA0(EEMPolyCplx *pNumerPoly, EEMSilarGrp *pDenomSilars,
  EEMSilar *pSilar, EEMComplex *pRes) {
    int		p = pSilar->power;
    pSilar->power = 0;
    _eemEvalPolySilar(pNumerPoly, pDenomSilars, &pSilar->loc, pRes);
    pSilar->power = p;
}

/**
  Silar.loc is the location of the singularity in the s or z plane.
  EEMPfe.Pole is the location of the pole in the s or z plane.
  If there are more zeros than poles, then the silar group is
  converted into a fraction of complex polynomials, and  
  complex polynomial division is performed until the order of numerator
  is less than that of the denominator.
  Heaviside's Expansion is used to determine the Gain of the EEMPfe.
  If there are multiple poles at the same location, then the partial
  fraction with the highest order in the denominator is obtained by
  Heavyside expansion.  The EEM_Pfe->Gain associated with the lower
  order EEM_Pfe (same pole) are determined by 
    A[i]=ith derivative ((s-p)^n * Y(s)) /factorial (n-i)
    evaluated at p
  where p is the location of the pole, n is the multiplicity of the
  pole, and the resulting partial fraction is of the form
    A[i]/[s-p]^i
**/
void 
eemSilarGrpToPfeGrp( EEMSilarGrp *pSG_in, EEMPfeGrp *pPG, TOPMask flags) {
    EEMSilarGrp		sg, *pSG = &sg;
    int			zeroPow, polePow;
    int			p, k;
    EEMComplex		Res;
    EEMPolyCplxRat	PR;
    double		kfacinv;
    TOPLogical		useRatPolyB;

    eemParamsCalc();
    pPG->shift = 0;			/* nuke origin component */
    eemPolyCplxZero( &pPG->poly, 0);	/* nuke FIR component */
    eemPfeGrpSetNum(pPG,0); 		/* nuke IIR component */
    eemSilarGrpGetPower(pSG_in, &zeroPow, &polePow);
    if ( polePow == 0 ) {
	/*
	 * There are no poles.  Just multiply out the zeros.
	 */
	eemSilarGrpToPolyCplx( pSG_in, &pPG->poly, 0);
	return;
    }
    eemSilarGrpConstructor( pSG);
    eemSilarGrpFlattenConj(pSG_in, pSG, 
      TOP_IsOn(flags,EEM_PFE_ExtractOrigin) ? &pPG->shift : NULL);

    useRatPolyB = zeroPow >= polePow;
    if ( useRatPolyB ) {
	/*
	 * For PFE to work, we must have a *proper* rational polynomial
	 * (more poles than zeros).  Steps below do this.
	 * After this operation, we will use the remainder polynomial
	 * to evaluate the numerator, but still use the
	 * poles of the silargroup to evaluate the denominator
	 */
    	eemPolyCplxRatConstructor(&PR);
        eemSilarGrpToPolyCplxRat(pSG, &PR);
	eemPolyCplxRatProperize(&PR);
	EEM_POLYCPLXRAT_Print(SPKG, &PR, "PFE Properized poly");
	eemPolyCplxCopy2(&PR.Whole, &pPG->poly);
	eemPolyCplxZero( &PR.Whole, 0);
    }

    EEM_SILAR_IDX_LOOP_FORW_BEGIN(pSG, silIdx, pSilar) {
         if ( (p=pSilar->power) >= 0 )
	     continue;
	 if ( useRatPolyB ) {
	     _eemEvalPolySilarA0( &PR.Numer, pSG, pSilar, &Res);
	 } else {
	     eemSilarGrpEvalA0(pSG, pSilar, &Res);
	 }
	 eemPfeGrpAppend( pPG, &Res, &pSilar->loc, p);

	if ( p != -1 ) {	    
	    /*
	     * For multiplicity, we are forced to convert to ratpoly form.
	     * Below, {rat} is H(s) * (s-p)^m where {p},{m} is the cur pole.
	     */
    	    EEMPolyCplxRat	rat;
    	    eemPolyCplxRatConstructor( &rat);

	    pSilar->power = 0;		/* NOTE: we are changing the pSG */
	    if ( useRatPolyB ) {
		/* We cant just copy the denom: must omit cur silar */
		eemPolyCplxCopy2( &PR.Numer, &rat.Numer);
		eemSilarGrpToPolyCplx( pSG, &rat.Denom, -1);
	    } else {
		eemSilarGrpToPolyCplxRat( pSG, &rat);
	    }

	    for (k=1; k<-p; k++) {
		eemPolyCplxRatDiff1( &rat);
		eemPolyCplxRatEval( &rat, &pSilar->loc, &Res);
		kfacinv = 1.0 / eemFactorial(k);
		Res.r *= kfacinv; Res.i *= kfacinv; 
		eemPfeGrpAppend( pPG, &Res, &pSilar->loc, p+k);
	    }
	    pSilar->power = p;		/* NOTE: restore silar power */
    	    eemPolyCplxRatDestructor( &rat);
	}
    } EEM_SILAR_IDX_LOOP_FORW_END();
    if ( useRatPolyB )
        eemPolyCplxRatDestructor( &PR);
    eemSilarGrpDestructor( pSG);
}


/**
  Converts a pfe to a silar with appropriate power
**/

void 
eemPfeToSilar( EEMPfe *pPfe, EEMSilar *pSilar) {
    pSilar->flags = 0;
    pSilar->loc = pPfe->pole;
    pSilar->power = pPfe->power;
}
  
/**
   Converts a PfeGrp to a SilarGrp.
   It is done by converting each Pfe to a fraction of complex
   polynomials; the resulting fractions all have a common denominator.
   Then the numerators are added.
   If {numzeros} is non-negative, then it specifies the number of zeros
   to expect in the numerator.  If negative, the algorithm is attempt
   to guess (and it will normally be wrong).
**/
void 
eemPfeGrpToSilarGrp( EEMPfeGrp *pPfeGrp, EEMSilarGrp *pSilGrp, int numzeros) {
    int		savepow, qidx;
    EEMPolyCplx	termpoly, numer;
    EEMSilar	*pKeySilar;

    eemParamsInit();

    eemPolyCplxConstructor(&termpoly);    
    eemPolyCplxConstructor(&numer);

    /*
     * Identify poles.  Go through all the terms of the pfegrp,
     * and find the maximum power pole at each location.  Create
     * a silar for each such one.
     */
    eemSilarGrpSetNum(pSilGrp,0);
    EEM_PFE_LOOP_BEGIN(pPfeGrp, pPfeTerm) {
        qidx = eemSilarGrpFindIdxByLoc(pSilGrp, &pPfeTerm->pole);
        if ( qidx >= 0 ) {
	    pKeySilar = &EEM_SILAR_Get(pSilGrp,qidx);
	    pKeySilar->power = TOP_MIN( pKeySilar->power, pPfeTerm->power);
        } else {
            eemPfeToSilar( pPfeTerm, &EEM_SILAR_Append(pSilGrp));        
	}
    } EEM_PFE_LOOP_END();

    /*
     * Create a numerator polynomial.  For each pfeterm, multiply
     * the gain by the silargrp equiv poly, ommitting the pfe's contribution
     * to that poly.  Add up all such polys to yield the numerator.
     */
    eemPolyCplxZero( &numer, 0);
    EEM_PFE_LOOP_BEGIN( pPfeGrp, pPfeTerm) {
        qidx = eemSilarGrpFindIdxByLoc(pSilGrp, &pPfeTerm->pole); 
	pKeySilar = &EEM_SILAR_Get(pSilGrp,qidx);
	savepow = pKeySilar->power;	/* save silar's power for restore */
	/* remove power from this pfe term */
	pKeySilar->power -= pPfeTerm->power;
	eemSilarGrpToPolyCplx( pSilGrp, &termpoly, -1);

	pKeySilar->power = savepow;	/* restore power */
        eemPolyCplxMulConstCplx2( &termpoly, &pPfeTerm->gain);
	eemPolyCplxMulX( &termpoly, pPfeTerm->numerpow);
	eemPolyCplxAdd2( &numer, &termpoly);
    } EEM_PFE_LOOP_END();

    /*
     * Add in the whole part.
     */
    if ( !eemPolyCplxZeroB( &pPfeGrp->poly) ) {
	eemSilarGrpToPolyCplx( pSilGrp, &termpoly, -1);
	eemPolyCplxMul2( &termpoly, &pPfeGrp->poly);
	eemPolyCplxAdd2( &numer, &termpoly);
    } else {
#ifdef notdef
	/* XXX: total hacked hueristic */
	for (k=EEM_POLYCPLX_Num(&numer)-1; k >= 1; k--) {
	    double magsq1 = EEM_CPLX_MAGSQ(&EEM_POLYCPLX_Get(&numer,k));
	    double magsq2 = EEM_CPLX_MAGSQ(&EEM_POLYCPLX_Get(&numer,k-1));
	    if ( magsq1 >= magsq2*2 )
		break;
	}
	eemPolyCplxSetNum( &numer, k+1);
#endif
	eemPolyCplxShorten( &numer);
    }
    EEM_POLYCPLX_Print(SPKG,&numer,"PfeToSilar Numer");
    if ( numzeros >= 0 && EEM_POLYCPLX_Num(&numer) > numzeros+1 ) {
	/* We only truncate the number of zeros here */
	eemPolyCplxSetNum( &numer, numzeros+1);
    }

    /*
     * Factor numer poly into zeros, add to silar grp.
     * SilarGrp Gain is leading coeficient of numerator.
     */
    eemPolyCplxToSilarGrp( &numer, &pSilGrp->gain, pSilGrp, 1);

    eemPolyCplxDestructor(&termpoly);
    eemPolyCplxDestructor(&numer);

    eemSilarGrpCompressConj(pSilGrp);
}

