/* 
    biquad.c

    Provides tcl procs from representing a silar rep
    into biquad configurations.

    Author: Kennard White

    Copyright (c) 1990-93 The Regents of the University of California.
    All rights reserved.  See the file "./copyright.h" for full copyright
    notice, limitation of liability, and disclaimer of warranty provisions.
#include "copyright.h"				COPYRIGHTENDKEY
*/
#if !defined(lint) && !defined(SABER)
static char rcsid[] = "$Header: /vol/cooley/cooley1/kennard/src/xpole/RCS/fltrRep.c,v 1.4 93/08/27 14:28:20 kennard Exp $";
#endif

#include "topFixup.h"
#include <stdio.h>
#include <tcl.h>

#include "topStd.h"
#include "topBuf.h"
#include "eemPoly.h"

#include "pole.h"
#include "filter.h"
#include "silar.h"

extern int ilogb( double x);
extern double scalbn( double x, int n);

/**
		FIR & IIR Cascade Structures

    The {cascade} filter structure logically consists of {sections} that are
    cascaded together.  The sections are may be FIR or IIR.  Nearly all
    quantities are treated internally as complex numbers; however,
    the options below may be used to generate real-valued filters.

    Options common to all structures:
	-real		Forces all values to be real (imaginary part zero).
			Generates warning message if not.
	-fmtstr		A eemCplxFmt() string used for formating all complex
			numbers.  Some format strings will cause the
			imaginary part to be omited if zero.

    Each filter section is described by an {XPCascSect} structure.  This
    contains copies of the silars that will be implemented by that section.
    (Note that copies must be used, because some silars must be split accross
    multiple sections).

    Each section contains two {EEMPolyCplx} elements: cooresponding
    to the zeros and poles.  FIR filter sections do not use the pole poly.
    The PolyCplx's are stored least-sig coef. first; this is backwards
    from how we want to print it out.  Thus in the calc stage we flip
    the coefs.  Must be careful not to use PolyCplx funcs. after this or
    weird results will occur.
**/

#define XPCASC_ZERO_IDX	0
#define XPCASC_POLE_IDX	1
typedef struct _XPCascSect XPCascSect;
struct _XPCascSect {
    int			SectNum;	/* for error reporting */
    EEMComplex		Gain;
    int			Shift;		/* in time: pos=forw=z^+1 */
    EEMSilarGrp		Sils;
    EEMPolyCplx		Taps[2];

    int			CurNumTaps[2];
    int			MinNumTaps[2];
    int			MaxNumTaps[2];
};

typedef struct _XPCascade XPCascade;
struct _XPCascade {
    TOPVector		Sects;		/* of XPCascSect */
    EEMComplex		Gain;
    int			Shift;		/* in time: pos=forw=z^+1 */
    TOPLogical		RealB;
    TOPLogical		FirB;
    char		*FmtStr;
    char		*WarnStr;
    int			MinNumSects;
    int			MaxNumSects;
    int			SectMinNumTaps[2];
    int			SectMaxNumTaps[2];
};


static void
_xpCascSectConstructor( XPCascSect *pSect, XPCascade *pCasc) {
    int	i;

    pSect->SectNum = -1;
    EEM_CPLX_SET( &pSect->Gain, 1.0, 0.0);
    pSect->Shift = 0;
    eemSilarGrpConstructor( &pSect->Sils);
    for (i=0; i < 2; i++) {
    	eemPolyCplxConstructor( &pSect->Taps[i]);
	pSect->CurNumTaps[i] = 0;
        pSect->MinNumTaps[i] = pCasc->SectMinNumTaps[i];
        pSect->MaxNumTaps[i] = pCasc->SectMaxNumTaps[i];
    }
}

static void
_xpCascSectDestructor( XPCascSect *pSect) {
    int	i;
    eemSilarGrpDestructor( &pSect->Sils);
    for (i=0; i < 2; i++) {
    	eemPolyCplxDestructor( &pSect->Taps[i]);
    }
}

static void
_xpCascConstructor( XPCascade *pCasc) {
    MEM_CLEAR(pCasc);
    EEM_CPLX_SET( &pCasc->Gain, 1.0, 0.0);
    pCasc->Shift = 0;
    pCasc->WarnStr = topBufLGet();
    topVecConstructor( &pCasc->Sects, sizeof(XPCascSect));
    pCasc->MinNumSects = 1;
    pCasc->MaxNumSects = -1;
    pCasc->SectMinNumTaps[0] = pCasc->SectMinNumTaps[1] = -1;
    pCasc->SectMaxNumTaps[0] = pCasc->SectMaxNumTaps[1] = -1;
}

static void
_xpCascDestructor( XPCascade *pCasc) {
    TOPVEC_LOOP_FORW_BEGIN( &pCasc->Sects, XPCascSect, pSect) {
	_xpCascSectDestructor( pSect);
    } TOPVEC_LOOP_FORW_END();
    topBufUnlock(pCasc->WarnStr);
    topVecDestructor( &pCasc->Sects);
}

static void
_xpCascSetNumSects( XPCascade *pCasc, int n) {
    int		k, orig = TOPVEC_Num(&pCasc->Sects);

    topVecSetNum( &pCasc->Sects, n);
    for ( k=n; k < orig; k++)
        _xpCascSectDestructor( &TOPVEC_Get(&pCasc->Sects,k,XPCascSect));
    for ( k=orig; k < n; k++)
        _xpCascSectConstructor( &TOPVEC_Get(&pCasc->Sects,k,XPCascSect), pCasc);
}

static XPCascSect*
_xpCascAppendSect( XPCascade *pCasc) {
    XPCascSect *pNewSect = &TOPVEC_Append( &pCasc->Sects, XPCascSect);
    _xpCascSectConstructor( pNewSect, pCasc);
    pNewSect->SectNum = TOPVEC_Num(&pCasc->Sects) - 1;
    return pNewSect;
}

/**
    Add {pSilar} to section {pSect}.  Only {tpow} of the powers of {pSilar}
    will be added.
**/
static void
_xpCascSectAddSilar( XPCascSect *pSect, EEMSilar *pSilar, int tpow) {
    int		typeIdx = pSilar->power > 0 ? XPCASC_ZERO_IDX : XPCASC_POLE_IDX;
    EEMSilar	*pNewSil;

    tpow = TOP_ABS(tpow);
    pNewSil = &EEM_SILAR_Append( &pSect->Sils);
    *pNewSil = *pSilar;
    pNewSil->power = pSilar->power > 0 ? tpow : - tpow;
    pSect->CurNumTaps[typeIdx] += tpow * (EEM_SILAR_ConjB(pSilar) ? 2 : 1);
}

/**
    Get (find or create) a section to contain some portion of {pSilar}.
    Note that {MaxNumTaps} specifies the number of *taps*, not silars.
**/
static XPCascSect*
_xpCascGetFree( XPCascade *pCasc, EEMSilar *pSilar, int *pPow) {
    int		typeIdx = pSilar->power > 0 ? XPCASC_ZERO_IDX : XPCASC_POLE_IDX;
    int		i, numFree, numNeed = TOP_ABS(*pPow);

    for ( i=0; i<2; i++) {
	TOPVEC_LOOP_BACK_BEGIN( &pCasc->Sects, XPCascSect, pSect) {
	    if ( pSect->MaxNumTaps[typeIdx] < 0 ) {
		*pPow = numNeed;
		return pSect;
	    }
	    numFree = pSect->MaxNumTaps[typeIdx] - pSect->CurNumTaps[typeIdx]-1;
	    if ( numFree <= 0 )	continue;
	    if ( ! EEM_SILAR_ConjB(pSilar) ) {
		*pPow = TOP_MIN( numNeed, numFree);
		return pSect;
	    } else {
		*pPow = TOP_MIN( numNeed, numFree/2);	/* auto rounding */
		if ( *pPow <= 0 )	continue;
		return pSect;
	    }
	} TOPVEC_LOOP_BACK_END();
        (void) _xpCascAppendSect( pCasc);
    }
    return NULL;
}

static void
_xpCascBuildSects( EEMSilarGrp *pSG, XPCascade *pCasc) {
    XPCascSect		*pSect;
    int			abspow, tpow;

    topVecSetNum( &pCasc->Sects, 0);

    EEM_SILAR_LOOP_BEGIN( pSG, pSilar) {
	if ( pSilar->power == 0 )		continue;
	if ( pCasc->FirB && pSilar->power < 0 )	continue;
	abspow = TOP_ABS(pSilar->power);
	for ( ; abspow > 0; ) {
	    tpow = abspow;
	    pSect = _xpCascGetFree( pCasc, pSilar, &tpow);
	    _xpCascSectAddSilar( pSect, pSilar, tpow);
	    abspow -= tpow;
	}
    } EEM_SILAR_LOOP_END();
}

#ifdef notdef
static void
_xpCascSectSetPairCoef( XPCascSect *pSect, int typeIndex) {
    XPSilar		**pSils = pSect->Silars[typeIndex];
    XPSilar		*pS = pSils[0];
    XPSilar		*pT = pSils[1];
    EEMComplex		*pCoefs = pSect->Coefs[typeIndex];

    /*IF*/ if ( (pS != NULL) && EEM_SILAR_ConjB(pS) ) {
	/*
	 * (z-x)(z-x~) = z*z -2*Re{x}*z + xx~, where xx~ is real 
	 * for x=(a+jb), xx~ = a*a + b*b = magsq(x)
	 */
	EEM_CPLX_SET( &pCoefs[0], 1.0, 0.0);
	EEM_CPLX_SET( &pCoefs[1], -2.0*pS->loc.r, 0.0);
	EEM_CPLX_SET( &pCoefs[2], EEM_CPLX_MAGSQ(&pS->loc), 0.0);
    } else if ( (pS != NULL) && (pT != NULL) ) {
	/*
	 * (z-x)(z-y) = z*z -(x+y)*z + x*y
	 */
	EEM_CPLX_SET( &pCoefs[0], 1.0, 0.0);
	eemCplxAdd3( &pS->loc, &pT->loc, &pCoefs[1]);
	eemCplxNeg1( &pCoefs[1]);
	eemCplxMul3( &pS->loc, &pT->loc, &pCoefs[2]);
    } else if ( pS != NULL ) {
	/*
	 * (z-x) = 0*z*z +z -x
	 * but we dont do that, and instead:
	 * z(z-x) = z*z -x*z +0
	 * XXX: This screws up the phase response!!!
	 */
	EEM_CPLX_SET( &pCoefs[0], 1.0, 0.0);
	EEM_CPLX_SET( &pCoefs[1], - pS->loc.r, - pS->loc.i);
	EEM_CPLX_SET( &pCoefs[2], 0.0, 0.0);
    } else {
	/*
	 * they are both null, just feed through 
	 */
	EEM_CPLX_SET( &pCoefs[0], 1.0, 0.0);
	EEM_CPLX_SET( &pCoefs[1], 0.0, 0.0);
	EEM_CPLX_SET( &pCoefs[2], 0.0, 0.0);
    }
}
#endif

static void
_xpCascCalcTaps( XPCascade *pCasc, XPCascSect *pSect, int typeIdx) {
    int			ntaps;
    EEMPolyCplx		*pTaps = &pSect->Taps[typeIdx];

    eemSilarGrpToPolyCplx( &pSect->Sils, pTaps, typeIdx==0 ? 1 : -1);

    ntaps =  EEM_POLYCPLX_Num(pTaps);
    if ( ntaps == 0 ) {
	/* default each section to unity pass through */
        eemPolyCplxOne(pTaps);
    } else if ( ntaps >= 2 ) {
        eemPolyCplxReverse( pTaps);
	pSect->Shift += typeIdx==0 ? ntaps-1 : 1-ntaps;
    }
    if ( pSect->MinNumTaps[typeIdx] > 0 ) {
        eemPolyCplxInsureNum( pTaps, pSect->MinNumTaps[typeIdx]);
    }
    if ( pSect->MaxNumTaps[typeIdx] > 0
      && EEM_POLYCPLX_Num(pTaps) > pSect->MaxNumTaps[typeIdx] ) {
	topBufAppendFmt(pCasc->WarnStr,"Too many taps (%d,%d)\n",
	  pSect->SectNum, typeIdx);
	eemPolyCplxSetNum( pTaps, pSect->MaxNumTaps[typeIdx]);
    }
}

static void
_xpCascCalcSects( XPCascade *pCasc) {
    TOPVEC_LOOP_FORW_BEGIN( &pCasc->Sects, XPCascSect, pSect) {
        _xpCascCalcTaps( pCasc, pSect, 0);
	if ( ! pCasc->FirB )
            _xpCascCalcTaps( pCasc, pSect, 1);
    } TOPVEC_LOOP_FORW_END();

    TOPVEC_LOOP_FORW_BEGIN( &pCasc->Sects, XPCascSect, pSect) {
	pCasc->Shift += pSect->Shift;
	pSect->Shift = 0;
    } TOPVEC_LOOP_FORW_END();
}

/*
	This is a mess b/c there are so many different structures
	and criterial for scaling:
	- Gabrial Cascade requires coef in (-1,1).  The coef
	  in each biquad must be scaled down the gabrial block requires
	  that all quads have the same scaling.  Each block has a left
	  shifter after it.  This structure allows for Cascade Form I,
	  and the dual-delay Cascade by scaling all numerator and all
	  denomenator coef except the first denom by a factor.
	  Scaling with the Cascade II structure of the alpha's cant be done 
	  in this framework b/c you would have to left-shift before the quad.
*/

#ifdef notdef
static void
_xpCascSectScale( XPCascade *pCasc) {
    int			c;
    double		m = 0, v;
    int			shifts;
    double		coef_scale;

    /* first pass: determine max value */
    TOPVEC_LOOP_FORW_BEGIN( &pCasc->Sects, XPCascSect, pSect) {
	for (c=0; c < 3; c++) {
	    if ( (v=fabs(pSect->Coefs[0][c].r)) > m)	m = v;
	    if ( (v=fabs(pSect->Coefs[1][c].r)) > m)	m = v;
	}
    } TOPVEC_LOOP_FORW_END();

    /* determine scale value 2^n */
    if ( m < 0.9999999 ) {
	shifts = 0;
	coef_scale = 1.0;
    } else {
    	shifts = 1 + ilogb(m);
	coef_scale = 1.0 / scalbn(1.0, shifts);
	
        TOPVEC_LOOP_FORW_BEGIN( &pCasc->Sects, XPCascSect, pSect) {
	    for (c=0; c < 3; c++) {
		pSect->Coefs[0][c].r *= coef_scale;
		pSect->Coefs[1][c].r *= coef_scale;
	    }
        } TOPVEC_LOOP_FORW_END();
    }
    pCasc->Shifts = shifts;
    pCasc->CoefScale = coef_scale;
}

#endif

static void
_xpCascSectResult( Tcl_Interp *ip, XPCascade *pCasc) {
    char		buf[200];

    sprintf( buf, "Cascade ");
    Tcl_SetResult(ip, buf, TCL_VOLATILE);
    eemCplxFmtTcl( ip, &pCasc->Gain, TOP_TRUE, pCasc->FmtStr);
    sprintf( buf, " %d ", pCasc->Shift);
    Tcl_AppendResult(ip,buf,NULL);

    if ( pCasc->RealB ) {
        TOPVEC_LOOP_FORW_BEGIN( &pCasc->Sects, XPCascSect, pSect) {
	   eemPolyCplxZeroImag( &pSect->Taps[0]);
	   eemPolyCplxZeroImag( &pSect->Taps[1]);
        } TOPVEC_LOOP_FORW_END();
    }
    TOPVEC_LOOP_FORW_BEGIN( &pCasc->Sects, XPCascSect, pSect) {
	Tcl_AppendResult(ip," { ",NULL);
        eemCplxFmtTcl( ip, &pSect->Gain, TOP_TRUE, pCasc->FmtStr);
        sprintf( buf, " %d ", pSect->Shift);
        Tcl_AppendResult(ip,buf,NULL);
	eemPolyCplxFmtTcl( ip, &pSect->Taps[0], TOP_TRUE, pCasc->FmtStr);
	if ( ! pCasc->FirB ) {
	    eemPolyCplxFmtTcl( ip, &pSect->Taps[1], TOP_TRUE, pCasc->FmtStr);
	}
	Tcl_AppendResult( ip, " } ",NULL);
    } TOPVEC_LOOP_FORW_END();
}

static int
_xpCascadeCmd( ClientData cld, Tcl_Interp *ip, int aC, char **aV) {
    char		*cmd, *viewspec;
    char		*tapsStr;
    int			opts, iVal, powPole, powZero;
    XPView		*pDM;
    XPSilarSnap		*pSR;
    EEMSilarGrp		sg;
    XPCascade		casc, *pCasc = &casc;

    _xpCascConstructor( pCasc);

    for ( opts=1; opts < aC; opts++) {
	cmd = aV[opts];
	if ( *cmd != '-' )
	    break;
	/*IF*/ if ( strcmp(cmd,"-real")==0 ) {
	    pCasc->RealB = TOP_TRUE;
	} else if ( strcmp(cmd,"-fmt")==0 ) {
	    pCasc->FmtStr = aV[++opts];
	} else if ( strcmp(cmd,"-fir")==0 ) {
	    pCasc->FirB = TOP_TRUE;
	} else if ( strcmp(cmd,"-minsects")==0 ) {
	    pCasc->MinNumSects = atoi(aV[++opts]);
	} else if ( strcmp(cmd,"-maxsects")==0 ) {
	    pCasc->MaxNumSects = atoi(aV[++opts]);
	} else if ( strcmp(cmd,"-numsects")==0 ) {
	    iVal = atoi(aV[++opts]);
	    pCasc->MinNumSects = pCasc->MaxNumSects = iVal;
	} else if ( strcmp(cmd,"-numtaps")==0 ) {
	    iVal = atoi(aV[++opts]);
    	    pCasc->SectMinNumTaps[0] = pCasc->SectMinNumTaps[1] = iVal;
    	    pCasc->SectMaxNumTaps[0] = pCasc->SectMaxNumTaps[1] = iVal;
	} else {
            Tcl_AppendResult(ip,aV[0],": bad option ``",cmd,"''", NULL);
	    return TCL_ERROR;
	}
    }

    if ( opts+1 != aC ) {
        Tcl_AppendResult(ip,"usage: ",aV[0]," ?-real? ?-fmt fmtstr?",
	  " viewspec", NULL);
	return TCL_ERROR;
    }
    viewspec = aV[opts];

    if ( (pDM=xpViewFindBySpec( viewspec, "SilarRep", NULL, TOP_TRUE)) == NULL ) {
        Tcl_AppendResult(ip,aV[0],": bad viewspec: ``",viewspec,"''",NULL);
	return TCL_ERROR;
    }
    pSR = xpSilarViewGetSnap( pDM, XPGS_Upd);
    eemSilarGrpConstructor( &sg);
    eemSilarGrpCopy( XP_SILSET_Grp(pSR), &sg);

    /*
     * Start actual filter synthysis here.
     */

    pCasc->Shift = eemSilarGrpExtractOrigin( &sg);
    eemSilarGrpGetPower( &sg, &powZero, &powPole);
    if ( pCasc->FirB && powPole != 0 ) {
	topBufAppendFmt(pCasc->WarnStr,"Filter has poles (%d), will ignore.\n",
	  powPole);
    }

    _xpCascBuildSects( &sg, pCasc);

    while ( TOPVEC_Num(&pCasc->Sects) < pCasc->MinNumSects ) {
	(void) _xpCascAppendSect( pCasc);
    }
    if ( pCasc->MaxNumSects > 0
      && TOPVEC_Num(&pCasc->Sects) > pCasc->MaxNumSects ) {
	topBufAppendFmt(pCasc->WarnStr,
	  "Filter has too many sections (%d), will ignore extra.\n",
	  TOPVEC_Num(&pCasc->Sects));
	_xpCascSetNumSects( pCasc, pCasc->MaxNumSects);
    }

    _xpCascCalcSects( pCasc);
    /* _xpCascSectScale( pCasc); */

    _xpCascSectResult( ip, pCasc);
    tapsStr = topBufLStrcpy( ip->result);
    Tcl_ResetResult(ip);
    Tcl_AppendElement( ip, pCasc->WarnStr, TOP_FALSE);
    Tcl_AppendElement( ip, tapsStr, TOP_FALSE);
    topBufUnlock( tapsStr);
    
    eemSilarGrpDestructor( &sg);
    _xpCascDestructor( pCasc);
    return TCL_OK;
}

void
xpFltrRepRegisterCmds( Tcl_Interp *ip) {
    Tcl_CreateCommand( ip, "cascade", _xpCascadeCmd, NULL, NULL);
}
