#!/usr/tools/bin/wish -f

#  A new version of the hierquery program -- updated to tcl6.6 & tk3.1
#  david richardson, normanb@citi.umich.edu
#
# This was originally a stand-alone tcl script by 
# David Richardson <normanb@citi.umich.edu>.  I have hacked it to make it
# more includable into xpole.  Hacks include:
# 1) eliminate auto-invoke at end of script
# 2) prefix all procs by "hq".
# 3) split hq.configinfo so it can be called directly (w/o tags)
# 4) modified configinfo to create parent if doesnt exist already

#  Globals at end.

#  Create a usable path name -- no .'s or spaces.
proc hq.getmash identifier {

    set identifier [string trim $identifier]
    regsub -all {[\.\ ]} $identifier "_" identifier
    return $identifier
}


#  Descend the path name tree and get a little info about each window.
proc hq.getinfo {viewer who self X} {
    global gY
    upvar #0 classes:[hq.getmash $who] lclasses
    set hpad 40
    set vpad 60

    set sY $gY
    set eY $gY
    if {[string length $who] > 0} {

	if {$self == "."} {
	    set name "."
	} else {
	    set name [send $who winfo name $self]
	}
	set class [send $who winfo class $self]
	set lclasses($class) 0

	# Now draw a box which will act like a button.
	set boxId [$viewer create rectangle [expr $X-20] [expr $sY-20] \
	    [expr $X+20] $sY -tags [list $class "Path=$self"]]
	# Can't depend on the "current" tag because the mouse will
	# be inside the rectangle, not actually on it.
	bind $viewer <1> [concat hq.configinfo [list $who] {[%W itemconfig \
	    [%W find closest [%W canvasx %x] [%W canvasy %y] ] -tag]} ]
	$viewer create text [expr $X+30] $sY -anchor sw -text "$name ($class)"

	set children [send $who winfo children $self]
	if {[llength $children] > 0} {
	    foreach child $children {
		set eY [incr gY $hpad]
		hq.getinfo $viewer $who $child [expr $X+$vpad]
	    }
	    $viewer create line $X $sY $X [expr $eY-10]
	}
	if {$self != "."} {
	    # And draw a line out to the parental line
	    $viewer create line [expr $X-20] [expr $sY-10] \
		[expr $X-$vpad] [expr $sY-10]
	}
    }
}


#  (Un)Highlight all of the boxes for widgets of a certain class.
#  Currently, this is called whenever a menu choice is (de)activated.
#  It should also be possible to do this with traces.
proc hq.filter {viewer classA} {
    upvar #0 $classA lclasses

    set classnames [array names lclasses]
    foreach class $classnames {
	set boxes [$viewer find withtag $class]
	if {$lclasses($class) == 1} {
	    foreach boxId $boxes {$viewer itemconfig $boxId -fill red}
	} else {
	    foreach boxId $boxes {$viewer itemconfig $boxId -fill {}}
	}
    }
}


#  A small amount of help information
proc hq.queryHelp {} {

    if [winfo exists .queryHelp] {
	# These protocols might not be available.
	catch {wm raise .queryHelp}
	catch {wm deiconify .queryHelp}
    } else {
    set helpWin [toplevel .queryHelp]
    set hText [text $helpWin.ins -height 5 -width 50 -export 0]
    $hText insert end "Scroll or scan through the widget hierarchy.\n"
    $hText insert end "Use the Edit menu to highlight widget classes.\n"
    $hText insert end "Click on a rectangle for configuration options.\n"
    # Turn off the widget so the text won't be altered
    $hText configure -state disabled
    button $helpWin.close -text "Close" -command "destroy $helpWin"
    pack append $helpWin $helpWin.ins {top expand} $helpWin.close {bottom fillx}
    }
}


#  Create the appropriate viewer for quering an interpreter.
#  For each application, we create a new toplevel named "interp:interp_name".
#  This can result in long, ugly paths, but hopefully we will avoid a collision.
proc hq.setupquery {who} {
    global gY
    upvar #0 classes:[hq.getmash $who] lclasses

    set queryWin ".[hq.getmash interp:$who]"
    # Make sure we don't already have a window open on this interpreter.
    # If so, destroy it -- consider this to be a 'rescan.'
    case $queryWin in [winfo children .] {destroy $queryWin} default {}
    toplevel $queryWin
    # This is kind of ugly, but it seems to allow menubar key traversal
    # More work should be done to clean up the use of focus.
    bind $queryWin <Any-Enter> "focus $queryWin"
    tk_bindForTraversal $queryWin

    set menuF [frame $queryWin.menus -relief sunken -bd 1]
	menubutton $menuF.file -text "File" -menu $menuF.file.m -underline 0
	menubutton $menuF.edit -text "Edit" -menu $menuF.edit.m -underline 0
	menubutton $menuF.help -text "Help" -menu $menuF.help.m -underline 0
	menu $menuF.file.m
	$menuF.file.m add command -label "Close" -command "destroy $queryWin"
	set editM [menu $menuF.edit.m -disabledforeground black]
	$editM add command -label "Highlight Class:" -command {} -state disabled
	menu $menuF.help.m
	$menuF.help.m add command -label "Help with query" -command {hq.queryHelp}
	pack append $menuF $menuF.file left $menuF.edit left $menuF.help right
	tk_menuBar $menuF $menuF.file $menuF.edit $menuF.help
    label $queryWin.title -text "Interpreter: $who"
    frame $queryWin.frame -bd 1
    pack append $queryWin $menuF {top fillx} $queryWin.title {top fill} \
	$queryWin.frame {top expand fill}

    scrollbar $queryWin.frame.vscroll -command "$queryWin.frame.viewer yview"
    scrollbar $queryWin.frame.hscroll -command "$queryWin.frame.viewer xview" \
	-orient horizontal
    set viewer [canvas $queryWin.frame.viewer -confine true -height 250 \
	-width 360 -yscroll "$queryWin.frame.vscroll set" \
	-xscroll "$queryWin.frame.hscroll set" -relief sunken]
    bind $viewer <Button-2> "$viewer scan mark %x %y"
    bind $viewer <B2-Motion> "$viewer scan dragto %x %y"
    pack append $queryWin.frame $queryWin.frame.hscroll {bottom fillx} \
	$queryWin.frame.vscroll {left filly} $viewer {right expand fill}
    wm minsize $queryWin 1 1
    wm title $queryWin "Interpreter: $who"

    $queryWin config -cursor watch
    # Need to reset gY each time we create a new viewer
    set gY 40
    hq.getinfo $viewer $who . 40
    # Pick a good size for the scrollregion.  Strange behaviour is possible
    # if the window size is larger than the scrollregion.
    set bbox [$viewer bbox all]
    set bX [expr [lindex $bbox 2]+20]
    set bY [expr [lindex $bbox 3]+20]
    $viewer config -scrollregion [list 0c 0c $bX $bY ]
    $queryWin config -cursor left_ptr
    # And set up the filter menu
    set classnames [lsort [array names lclasses]]
    foreach i $classnames {
	$editM add checkbutton -label $i -onvalue 1 -offvalue 0 \
	-variable classes:[hq.getmash $who]($i) \
	-command "hq.filter $viewer classes:[hq.getmash $who]"  
    }
}


#  Get and manipulate the configuration options for a specific window.
proc hq.configinfowin {who widget} {
    if {$widget == {} || $widget == "."} return
    set class [send $who winfo class $widget]

    # Ugly pathname, but does the right thing when the viewer is closed.
    set pntWin .[hq.getmash interp:$who]
    if { ![winfo exist $pntWin] } {
	hq.setupquery $who
    }
    set configWin "$pntWin.config:[hq.getmash $widget]"
    catch {destroy $configWin}
    toplevel $configWin -class ConfigInfo -geom 200x100

    set titleL [label $configWin.title -text "Configuration: $widget ($class)"]
    set quitB [button $configWin.quit -text "Close" \
	-command "destroy $configWin"]
    # Is there a better way to construct a message widget?
    set instruct [message $configWin.ins -text "Enter new information and press RETURN to configure.  Double click on the Value field to copy or edit the old value." -aspect 800]
    set configF [frame $configWin.configinfo -bd 1]
    set configT [frame $configF.configTitle -bd 2 -relief raised]
    set opt [label $configT.opt -text [format "%-20s" "Option"] \
	-relief raised -padx 2]
    set val [label $configT.val -text [format "%-10s" "Value"] \
	-relief raised -padx 2]
    set newVal [label $configT.new -text "New Value" \
	-relief raised -padx 2]
    pack append $configT $opt {left} $val {left} $newVal {left expand fillx}
    pack append $configF $configT {top fillx}

    # For each config option, we create:
    #	datumF	a frame using the option name
    #	opt	a label for the name
    #	val	a label for the current value
    #	newVal	an entry to hold a new value

    set information [send $who $widget config]
    #option add *ConfigInfo.configinfo.Frame*relief raised
    foreach datum $information {
	if {[llength $datum] == 5} {
	    set name [string range [lindex $datum 0] 1 end]
	    set value [lindex $datum 4]
	    set datumF [frame $configF.$name -bd 2]
	    set opt [label $datumF.opt -text [format "%-20s" $name] \
		-relief raised -padx 2]
	    set val [label $datumF.val  -text [format "%-10s" $value] \
		-relief raised -padx 2]
	    set newVal [entry $datumF.new -relief raised]
	    # Some error checking should be done here.  At least a catch
	    # for non-existent window, unsuported option, etc.  Would be
	    # nice to do boundry/option type checking on this end, but that
	    # is a little too ambitious.  For now, errors appear on stdout.
	    bind $newVal <Return> [concat send [list $who] $widget config \
		-$name {[list [%W get]]}]
	    bind $val <Double-1> "$newVal delete 0 end; $newVal insert 0 \
		[list $value]"
	    pack append $datumF $opt left $val left $newVal {left expand fillx}
	    pack append $configF $datumF {top fillx}
	} else {}
    }

    pack append $configWin $titleL {top fillx} $instruct {top fillx} \
	$configF {top expand fill} $quitB {bottom fillx}

    wm minsize $configWin 1 1
    wm title $configWin "Configuration: $widget"

}


#  Get and manipulate the configuration options for a specific window.
proc hq.configinfo {who tags} {
    # Need to do some parsing of the tags field to find out who we're working
    # on.  Expect tags to be of the form: {-tags {} {} {} {blah Path=.window}}
    set path [lindex $tags 4]
    set path [lindex $path [lsearch $path "Path=.*"] ]
    set widget [string range $path [string first . $path] end]
    hq.configinfowin $who $widget
}

#  Get the list of the current interpreters.  Taken from Ousterhout's
#  demos/rmt application.
proc hq.fillAppsMenu m {

    catch {$m delete 0 last}
    foreach i [lsort [winfo interps]] {
	$m add command -label $i -command [list hq.setupquery $i]
    }
}


proc hq.about {{useTop ""}} {
    global HqVersion

    if [winfo exists $useTop.about] {
	catch {wm raise $useTop.about}
	catch {wm deiconify $useTop.about}
    } else {
    set aboutF [toplevel $useTop.about]
    set titleL [label $aboutF.title -text "hierQuery: $HqVersion"]
    set closeB [button $aboutF.close -text "Close" -command "destroy $aboutF"]
    pack append $aboutF $titleL {top fillx} $closeB {bottom fillx}
    }
}


#  The equivalent of main().
#  May be called with any toplevel widget, or just use . by default
#  Should be possible to incorporate this into other scripts/applications.
proc hierQuery {{top "."}} {

    if {$top != "."} {
	set useTop $top
    } else { set useTop "" }

    set appF [frame $useTop.apps -relief sunken -bd 1]

    menubutton $appF.file -text "File" -menu $appF.file.m -underline 0 \
	-width 10 -height 2 -relief raised
    menubutton $appF.edit -text "Edit" -menu $appF.edit.m -underline 0 \
	-state disabled -width 10 -height 2 -relief raised
    menubutton $appF.help -text "Help" -menu $appF.help.m -underline 0 \
	-width 10 -height 2 -relief raised

    set fileM [menu $appF.file.m]
    $fileM add cascade -label "Select Application" \
	-command "hq.fillAppsMenu $fileM.apps" \
	-menu $fileM.apps -underline 0 -accelerator =>
    menu $fileM.apps
    $fileM add command -label "Quit" -command "destroy $top" -underline 0

    menu $appF.edit.m
    [menu $appF.help.m] add command -label "About hierQuery" -underline 0 \
	-command "hq.about $useTop"

    pack append $appF $appF.file left $appF.edit left $appF.help right
    tk_menuBar $appF $appF.file $appF.edit $appF.help

    pack append $top $appF {top expand fillx}
    wm title $top hierQuery
}


#  Globals:

    # An array of form "classes:$interpName" for each interp used for filtering
    set HqVersion "\$Id: hierQuery.tcl,v 1.1 93/03/22 14:26:09 kennard Exp Locker: kennard $"
    set gY 40

# Should be a mono/constant spaced font
option add *ConfigInfo.configinfo*Label.font fixed

proc hqBoot {} {
    global hqEmbed
    if { [info exist hqEmbed] } return
    hierQuery
}
hqBoot
